# -*- coding: utf-8 -*-
from path import path
import ped_parser
import yaml

from pymip.compat import iteritems


class Family(object):

    """Interface to a MIP family.

    Responsibilities include: starting analyses, etc.

    Args:
        base_dir (path): root level path to the top level family dir
        load (bool): initialize loading automatically [default: True]
    """

    def __init__(self, base_dir, load=True):
        super(Family, self).__init__()
        self.base_dir = path(base_dir)
        self.family_id = self.base_dir.basename()

        self.config_path = self.base_dir.joinpath("{}_config.yaml"
                                                  .format(self.family_id))
        self.ped_path = self.base_dir.joinpath("{}_pedigree.txt"
                                               .format(self.family_id))
        self.sampleinfo_path = self.base_dir.joinpath("{}_qc_sampleInfo.yaml"
                                                      .format(self.family_id))

        self.config = {}
        self.ped = None
        self._family = {}
        self._samples = {}

        if load:
            self.load()

    def load(self):
        """Load the class with data from the file system."""
        # parse qc sample info data
        self._load_sampleinfo()

        # load analysis config data
        self._load_config()

        # setup pedigree
        self._load_pedigree()

    @property
    def owner(self):
        """Return the first institute, always >= 1 institute."""
        return self._institutes[0]

    @property
    def collaborators(self):
        """Return all but the first institute (owner)."""
        return self._institutes[1:]

    @property
    def analyzed_at(self):
        """Return the date of when the analysis was run.

        Returns:
            datetime: timestamp for when the analysis was run
        """
        # N.B. already parsed as datetime object by PyYAML
        return self._family['AnalysisDate']

    @property
    def is_complete(self):
        """Check the analysis run status of the family."""
        analysis_run_status = self._family.get('AnalysisRunStatus')

        if analysis_run_status == 'Finished':
            return True

        elif analysis_run_status == 'notFinished':
            return False

    @property
    def is_wgs(self):
        """Check if the analysis is whole genome."""
        # infer this from the absolute path to the sample info file
        return True if 'genomes' in self.base_dir.abspath() else False

    @property
    def human_genome_build(self):
        """Return the human genome reference build."""
        return self._family['HumanGenomeBuild']['Source']

    @property
    def human_genome_version(self):
        """Return the version of the human genome reference build."""
        return self._family['HumanGenomeBuild']['Version']

    @property
    def ready_vcf(self):
        """Return the path to the most complete VCF file."""
        return self._family['VCFFile']['ReadyVcf']['Path']

    @property
    def clinical_vcf(self):
        """Return the path to the clinical VCF file."""
        return self._family['VCFFile']['Clinical']['Path']

    @property
    def research_vcf(self):
        """Return the path to the research VCF file."""
        return self._family['VCFFile']['Research']['Path']

    def gene_lists(self):
        """Return meta data on all gene lists."""
        included_lists = self._family['VCFParser']['SelectFile']['Database']

        return {data['Acronym']: {'name': data['Acronym'],
                                  'version': data['Version'],
                                  'full_name': data.get('CompleteName',
                                                        data['Acronym']),
                                  'date': data.get('Date')}
                for _, data in iteritems(included_lists)}

    def default_genelist_ids(self, ped_key='Clinical_db'):
        """Return a simple list of ids for the default gene lists."""
        ind_lists = (data.extra_info.get(ped_key, '').split(';')
                     for ind_id, data in iteritems(self.ped.individuals))

        # flatten list of lists
        return set(item for sublist in ind_lists for item in sublist)

    def default_genelists(self):
        """Return a simple list of default gene lists."""
        genelist_ids = self.default_genelist_ids()

        return [data for list_id, data in iteritems(self.gene_lists())
                if list_id in genelist_ids]

    def genelist_name(self):
        """Generate stringified caption for default gene lists."""
        default_genelists = self.default_genelists()
        assert len(default_genelists) > 0, 'No default gene lists'

        gene_lists = ("{} ({})".format(gene_list.get('full_name'),
                                       gene_list.get('version'))
                      for gene_list in default_genelists)

        return ' + '.join(gene_lists)

    @property
    def genelist_path(self):
        """Return the master gene list template."""
        reference_dir = path(self.config['referencesDir'])
        template_file = self._family['VCFParser']['SelectFile']['File']

        return reference_dir.joinpath(template_file)

    def chanjo_sample_outputs(self):
        """Return a list of chanjo coverage outputs with samples ids."""
        samples = ((sample_id, data['Program']['ChanjoAnnotate'].values())
                   for sample_id, data in iteritems(self._samples))

        # expect only one chanjo output per sample
        samples_flat = ((sample_id, data[0]) for sample_id, data in samples)

        return [(sample_id, path(data['Bed']['Path']))
                for sample_id, data in samples_flat]

    def chanjo_outputs(self):
        """Return a list of chanjo coverage outputs."""
        return [cov_path for _, cov_path in self.chanjo_sample_outputs()]

    @property
    def pedigree_svgpath(self):
        """Return the path to the Madeline SVG pedigree output."""
        return self._program('Madeline').get('Path')

    def scout_config_clinical(self):
        """Build a clinical Scout config file."""
        base_config = self._scout_config()
        base_config['load_vcf'] = self.clinical_vcf

        return base_config

    def scout_config_research(self):
        """Build a research Scout config."""
        base_config = self._scout_config()
        base_config['load_vcf'] = self.research_vcf

        return base_config

    @property
    def _institutes(self):
        """Return all institutes."""
        return self._family.get('InstanceTag', [])

    def _program(self, program_id, default=None):
        return self._family['Program'].get(program_id, default or {})

    def _load_sampleinfo(self):
        """Load the QC sample info file using YAML parser."""
        try:
            # open and read the (existing) QC metrics file
            with self.sampleinfo_path.open('r') as handle:
                data = yaml.load(handle)

        except IOError:
            # sample info file didn't exist as expected
            error_message = ("QC sample info file not found for {}"
                             .format(self.family_id))
            raise IOError(error_message)

        # only care about loading a subtree of the data
        data_flat = data[self.family_id]
        self._family = data_flat[self.family_id]

        # remove family information subtree
        del data_flat[self.family_id]

        # load the sample specific information
        self._samples = data_flat

    def _load_config(self):
        """Load the MIP config as YAML file."""
        try:
            with self.config_path.open('r') as handle:
                # load the config file as YAML
                self.config = yaml.load(handle)

        except IOError:
            # config file didn't exist as expected
            error_message = ("Config file not found for {this.family_id}"
                             .format(this=self))
            raise IOError(error_message)

    def _load_pedigree(self):
        """Load the pedigree using ``ped-parser``."""
        try:
            # load the pedigree information into ped-parser
            with path(self.ped_path).open('r') as handle:
                self.ped = ped_parser.FamilyParser(handle, family_type='alt',
                                                   cmms_check=False)
        except IOError:
            # pedigree file didn't exist as expected
            error_message = ("Pedigree file not found for {this.family_id}"
                             .format(this=self))
            raise IOError(error_message)

    def _scout_config(self):
        """Build a base Scout config."""
        individuals = {sample_id: {
            'name': sample_id,
            'bam_path': sample['MostCompleteBAM']['Path'],
            'capture_kit': sample.get('Capture_kit', []),
        } for sample_id, sample in iteritems(self._samples)}

        return {
            'load': True,
            'ped': self._family['PedigreeFile']['Path'],
            'igv_vcf': self.ready_vcf,
            'human_genome_build': self.human_genome_build,
            'human_genome_version': self.human_genome_version,
            'owner': self.owner,
            'collaborators': self.collaborators,
            'madeline': self.pedigree_svgpath,
            'coverage_report': None,
            'gene_lists': self.gene_lists(),
            'default_gene_lists': list(self.default_genelist_ids()),
            'individuals': individuals,
        }



# -*- coding: utf-8 -*-
from path import path


class MIP(object):

    """Wrapper for control class for MIP pipeline.

    Args:
        base_path (path): path to root of MIP setup
    """

    def __init__(self, base_path):
        super(MIP, self).__init__()
        self.base_path = path(base_path)



# -*- coding: utf-8 -*-
from pymip.utils import fastq_name


def test_fastq_name():
    """Test FASTQ file name generator."""
    name = fastq_name('1', '140127', 'H8A7MADXX', '0002P021', 'ACTTGA', '2')
    expected = '1_140127_H8A7MADXX_0002P021_ACTTGA_2.fastq.gz'

    assert name == expected



# -*- coding: utf-8 -*-
from datetime import date
import logging
import os.path

LEVELS = {0: 'ERROR', 1: 'WARNING', 2: 'INFO', 3: 'DEBUG'}
TEMPLATE = "[%(asctime)s] %(name)-25s %(levelname)-8s %(message)s"
FORMATTER = logging.Formatter(TEMPLATE)


def setup_daily_logging(log_dir, level='INFO'):
    """Setup logging using ``FileHandler`` based on current date."""
    # use isoformatted date as filename: "2015-11-14"
    today = date.today().isoformat()
    filename = os.path.join(log_dir, "{}.log".format(today))

    # setup handler that will rotate daily
    daily_handler = logging.FileHandler(filename)

    # configure handler
    daily_handler.setFormatter(FORMATTER)
    daily_handler.setLevel(level)

    return daily_handler


def setup_stream_logging(level='ERROR'):
    """Setup logging to STDERR."""
    # add a handler to the logger
    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(FORMATTER)

    return console_handler


def setup_logging(filename, file_level='INFO', stderr_level='ERROR',
                  root_level='DEBUG'):
    """Configure central logging for the package."""
    # get the root logger
    root_logger = logging.getLogger()
    root_logger.setLevel(root_level)

    # setup stream logging to STDERR
    console = setup_stream_logging(level=stderr_level)
    file_handler = setup_daily_logging(filename, level=file_level)

    root_logger.addHandler(console)
    root_logger.addHandler(file_handler)
    return root_logger



# -*- coding: utf-8 -*-
"""
pymip
~~~~~~~~~~~~~~~~~~~

Provides a Python API to manipulate MIP

:copyright: (c) 2015 by Robin Andeer
:licence: MIT, see LICENCE for more details
"""
import logging
from pkg_resources import get_distribution

# Generate your own AsciiArt at:
# patorjk.com/software/taag/#f=Calvin%20S&t=PyMIP
__banner__ = r"""
╔═╗┬ ┬╔╦╗╦╔═╗
╠═╝└┬┘║║║║╠═╝
╩   ┴ ╩ ╩╩╩    by Robin Andeer
"""

__title__ = 'pymip'
__summary__ = 'Provides a Python API to manipulate MIP'
__uri__ = 'https://github.com/robinandeer/pymip'

__version__ = get_distribution(__title__).version

__author__ = 'Robin Andeer'
__email__ = 'robin.andeer@gmail.com'

__license__ = 'MIT'
__copyright__ = 'Copyright 2015 Robin Andeer'

# the user should dictate what happens when a logging event occurs
logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
from path import path

from pymip.api import Family


class Institute(object):

    """docstring for Institute"""

    def __init__(self, base_path, cust_id):
        super(Institute, self).__init__()
        self.cust_id = cust_id
        self.base_path = path(base_path)
        self.families = {}

    def load_family(self, family_id):
        """Load a family from the instutute.

        Args:
            family_id (str): id of family

        Returns:
            Family: initialized family object
        """
        family_dir = self.base_path.joinpath(family_id)
        # cache the family in the class
        self.families[family_id] = Family(base_dir=family_dir)

        return self.families[family_id]



# -*- coding: utf-8 -*-
from __future__ import division

from path import path
import yaml

from pymip.compat import iteritems


class FamilyOutput(object):

    """docstring for FamilyOutput"""

    def __init__(self, base_path, load=False):
        super(FamilyOutput, self).__init__()
        self.base_path = path(base_path)
        self.family_id = None
        self._data = None
        self.samples = []

        if load:
            self.load()

    def load(self):
        """Load an analysis family into the class object."""
        # extract the name of the family directory
        family_id = self.base_path.basename()
        self.family_id = int(family_id) if family_id.isdigit() else family_id

        # read in the YAML QC metrics file
        files = [file_path for file_path in self.base_path.listdir()
                 if file_path.endswith('_qcmetrics.yaml')]

        try:
            qc_file = files[0]
        except IndexError:
            # QC metrics file didn't exist as expected
            raise IOError("QC metrics file not found: {}"
                          .format(self.base_path))

        # open and read the (existing) QC metrics file
        with qc_file.open() as handle:
            data = yaml.load(handle).values()[0]

        # do some initial parsing to lose superfulous data
        for key, values in iteritems(data):
            if key == self.family_id:
                # store QC data subtree
                self._data = values

            else:
                # add a sample output instance
                self.samples.append(SampleOutput(key, data=values))


class SampleOutput(object):

    """docstring for SampleOutput"""

    def __init__(self, sample_id, data=None):
        super(SampleOutput, self).__init__()
        self.sample_id = sample_id
        self._raw = data

        # store the subtree with actually relevant data ("all lanes")
        for key, value in iteritems(data):
            if 'lanes' in key:
                self._data = value

    @property
    def total_reads(self):
        """Return the total number of reads."""
        return (self._data['CalculateHsMetrics']['Header']['Data']
                          ['TOTAL_READS'])

    @property
    def total_reads_m(self):
        """Return the total number of reads."""
        return self.total_reads / 1000000

    @property
    def duplicates(self):
        """Return the duplicates in percent."""
        return (self._data['MarkDuplicates']['Header']['Data']
                          ['PERCENT_DUPLICATION'])

    @property
    def mean_target_coverage(self):
        """Return mean target coverage."""
        return (self._data['CalculateHsMetrics']['Header']['Data']
                          ['MEAN_TARGET_COVERAGE'])

    @property
    def off_bait(self):
        """Return off bait in percent."""
        return (self._data['CalculateHsMetrics']['Header']['Data']
                          ['PCT_OFF_BAIT'])

    @property
    def aligned_reads(self):
        """Return the pass filter unique align reads."""
        count = (self._data['CalculateHsMetrics']['Header']['Data']
                           ['PF_UQ_READS_ALIGNED'])
        return count / self.total_reads

    @property
    def ti_tv(self):
        """Return Ti/Tv ratio."""
        return (self._data['CalculateHsMetrics']['Header']['Data']
                           ['PF_UQ_READS_ALIGNED'])

    @property
    def gender(self):
        """Return the predicted gender by chanjo."""
        return self._data['ChanjoSexCheck']['gender']

    def target_coverage(self, level=10):
        """Return target coverage at different levels."""
        return (self._data['CalculateHsMetrics']['Header']['Data']
                          ["PCT_TARGET_BASES_{}X".format(level)])




# -*- coding: utf-8 -*-
import logging

import click

from pymip import __version__
from pymip.log import LEVELS, setup_logging
from pymip.scripts.utils import parse_config
from pymip.utils import start_analysis

logger = logging.getLogger(__name__)


@click.group()
@click.option('-c', '--config', default='~/.clinical/config.yaml',
              type=click.Path(exists=True), help='path to config file')
@click.option('-l', '--log-dir', default='/tmp', type=click.Path(exists=True),
              help='path to log file directory')
@click.option('-v', '--verbose', count=True)
@click.version_option(__version__)
@click.pass_context
def cli(ctx, config, log_dir, verbose):
    """MIP and downstream processing of data at Clinical Genomics."""
    # setup logging handlers
    setup_logging(log_dir, stderr_level=LEVELS.get(verbose))

    # read in config values
    ctx.obj = parse_config(config)

    logger.info("Running v%s", __version__)
    ctx.obj['mongo_db'] = ctx.obj.get('mongo_db', 'scoutProd')
    ctx.obj['mongo_port'] = ctx.obj.get('mongo_port', 27017)
    ctx.obj['mongo_host'] = ctx.obj.get('mongo_host', 'localhost')
    ctx.obj['madeline_exe'] = ctx.obj.get('madeline_exe', '/usr/bin/madeline2')


@cli.command()
@click.argument('family', type=str)
@click.option('-cc', '--clusterconstant', type=click.Path(exists=True))
@click.option('-c', '--config', type=click.Path(exists=True))
@click.option('-g', '--gene-list', type=str)
@click.option('-m', '--mip', type=click.Path(exists=True))
@click.pass_context
def start(ctx, family, clusterconstant, config, gene_list, mip):
    """Start a new MIP analysis for a family."""
    return_code = start_analysis(family, clusterconstant, config,
                                 gene_list=gene_list, mip_path=mip)



# -*- coding: utf-8 -*-
import itertools
import os
import subprocess


def fastq_name(lane, date, flowcell, sample, index, read):
    """Build a string representing FASTQ file name.

    The naming follows MIP conventions. The syntax is as follows:
        1_140818_H9FD3ADXX_000127T_GATCAG_1.fastq.gz
    """
    if read not in ('1', '2'):
        raise ValueError("'read' must be either of directions: '1' or '2'")

    elif '_' in sample:
        raise ValueError("'sample' must not contain any underscores")

    return ("{lane}_{date}_{flowcell}_{sample}_{index}_{read}.fastq.gz"
            .format(lane=lane, date=date, flowcell=flowcell, sample=sample,
                    index=index, read=read))


def crawl_files(root_path, end_patten='_qc_sampleInfo.yaml'):
    """Crawl for files matching a certain end pattern.

    Args:
        root_path (str): where to start crawling from and down
        end_patten (str, optional): pattern to match files against

    Yields:
        str: path to files matching the end pattern
    """
    # crawl all files in the directory tree, building full paths, nested
    all_files = ((os.path.join(root, a_file) for a_file in files)
                 for root, _, files in os.walk(root_path, followlinks=True))

    # un-nest the list of lists
    all_files_flat = itertools.chain.from_iterable(all_files)

    # match paths against the provided pattern
    return (a_file for a_file in all_files_flat if a_file.endswith(end_patten))


def start_analysis(family_id, clusterconstant_path, config_path,
                   gene_list=None, mip_path=None):
    """Lunch a new MIP analysis for a specific family.

    Args:
        family_id (str): unique id for the family to analyze
        clusterconstant_path (path): root path for the analysis
        config_path (path): absolute path to the MIP config file
        gene_list (str, optional): name of the gene list file
        mip_path (path, optional): path to MIP script file

    Rerurns:
        int: return code from the executed process
    """
    command = []

    # configure the executable
    if mip_path:
        command.append('perl')
        command.append(mip_path)
    else:
        command.append('mip.pl')

    # add the family option
    command.append('--familyID')
    command.append(family_id)

    # add YAML config file path
    command.append('--configFile')
    command.append(config_path)

    # configure the "cluster constant path"
    command.append('--clusterConstantPath')
    command.append(clusterconstant_path)

    if gene_list:
        command.append('--vcfParserSelectFile')
        command.append(gene_list)

    return subprocess.Popen(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
